# -*- coding: utf-8 -*-

import os
import glob
from collections import OrderedDict, namedtuple
from . import config


postrender = OrderedDict()
pathgen = OrderedDict()
integration = OrderedDict()
extension = OrderedDict()

PathGenerator = namedtuple('PathGenerator', ['name', 'handler'])
PostRender = namedtuple('PostRender', ['name', 'handler', 'default'])
Extension = namedtuple('Extension', ['name', 'ext', 'handler', 'options'])


def register_postrender(name, handler, default=None):
    '''Add postrender function to registry

    :param name: Name of postrender function (used as label in ui)
    :param handler: Postrender function
    '''

    postrender[name] = PostRender(name, handler, default)


def unregister_postrender(name):
    '''Remove postrender function from registry'''

    postrender.pop(name, None)


def register_path(name, handler):
    '''Add a path generator function to registry

    :param name: Name of path generator (used as label in ui)
    :param handler: Path generator function
    '''

    pathgen[name] = PathGenerator(name, handler)


def unregister_path(name):
    '''Remove path generator function from registry

    :param name: Name of path generator function
    '''

    pathgen.pop(name, None)


def register_integration(name, obj):
    '''Add a path generator function to registry

    :param name: Name of Integration
    :param obj: Integration obj
    '''

    integration[name] = obj


def unregister_integration(name):
    '''Remove path generator function from registry

    :param name: Name of integration to remove
    '''

    integration.pop(name, None)


def register_extension(name, ext, handler, options=None):
    '''Register a function to handle playblasting a specific file extension

    :param name: Nice name of the extension
    :param ext: File extension
    :param handler: Handler function that performs the playblasting
    '''

    extension[name] = Extension(name, ext, handler, options)


def unregister_extension(name):
    '''Register a function to handle playblasting a specific file extension

    :param name: Nice name of the extension to unregister
    '''

    extension.pop(name, None)


def init():
    '''Discover presets'''

    postrender.clear()
    pathgen.clear()
    integration.clear()
    extension.clear()

    # Find registered presets in MVP_PRESETS path
    for path in config.PRESETS_PATH:

        # Import all python files on MVP_PRESETS path
        for f in glob.glob(os.path.join(path, '*.py')):
            base = os.path.basename(f)
            name = os.path.splitext(base)[0]
            __import__(name)

        # Import all packages on MVP_PRESETS path
        for f in glob.glob(os.path.join(path, '*', '__init__.py')):
            name = os.path.basename(os.path.dirname(f))
            __import__(name)

    # Import builtin extensions
    from . import extensions



# -*- coding: utf-8 -*-

import json
import glob
import os
from . import config


def get_presets():
    '''Get a generator yielding preset name, data pairs'''

    for path in config.PRESETS_PATH:
        for f in glob.glob(os.path.join(path, '*.json')):

            base = os.path.basename(f)
            name = os.path.splitext(base)[0]

            with open(f, 'r') as f:
                data = json.loads(f.read())

            yield name, data


def get_preset(name):
    '''Get a preset by name'''

    for n, s in get_presets():
        if name == n:
            return s


def find_preset(name):
    '''Find the path to a given preset...'''

    for path in config.PRESETS_PATH:
        prospect = os.path.join(path, name + '.json')
        if os.path.isfile(prospect):
            return prospect

    raise ValueError('Could not find a preset named %s', name)


def new_preset(name, data):
    '''Create a new preset from viewport state data

    :param name: Name of the preset
    :param data: Viewport state dict

    usage::

        import mvp
        active = mvp.Viewport.active()
        mvp.new_preset('NewPreset1', active.get_state())
    '''

    preset_path = os.path.join(config.PRESETS_PATH[0], name + '.json')
    with open(preset_path, 'w') as f:
        f.write(json.dumps(data))


def del_preset(name):

    preset_path = find_preset(name)
    if os.path.exists(preset_path):
        os.remove(preset_path)



'''
Standard Validators
===================
'''
import re
from .exc import ValidationError


def regex(rstr, msg='Does not match regex'):
    r = re.compile(rstr)
    def check_string(value):
        match = r.search(value)
        if not match:
            raise ValidationError(msg)
        return True
    return check_string


def checked(value):
    if not value:
        raise ValidationError('Must be checked')
    return True


def email(value):
    r = re.compile(r'^[\w\d!#$%^&*(){\-_}|]+@[\w\d\-_]+[.][a-z]{2,4}')
    match = r.search(value)
    if not match:
        raise ValidationError('Not a valid email address')
    return True


def required(value):
    if not bool(value):
        raise ValidationError('Missing required field')
    return True


def min_length(num_characters):

    def check_length(value):
        if len(value) < num_characters:
            raise ValidationError('Min Length {0}'.format(num_characters))
    return check_length


def max_length(num_characters):

    def check_length(value):
        if len(value) > num_characters:
            raise ValidationError('Max Length {0}'.format(num_characters))
    return check_length




# -*- coding: utf-8 -*-
from contextlib import contextmanager

import maya.app.renderSetup.model.renderSetup as renderSetup
from maya import cmds


@contextmanager
def enabled_render_layers():
    old_layer = cmds.editRenderLayerGlobals(
        query=True,
        currentRenderLayer=True,
    )
    try:
        rs = renderSetup.instance()

        def switchToLayer(layer):
            def _switch():
                rs.switchToLayer(layer)
            return _switch

        enabled_layers = []
        for layer in rs.getRenderLayers() + [rs.getDefaultRenderLayer()]:

            layer.switchToLayer = switchToLayer(layer)
            if layer.isRenderable():
                enabled_layers.append(layer)

        yield enabled_layers
    finally:
        cmds.editRenderLayerGlobals(currentRenderLayer=old_layer)



# -*- coding: utf-8 -*-
from .exc import FieldNotInstantiated
from . import controls
from .utils import Ordered
from .validators import required
from copy import deepcopy


missing = object()


def get_key(*args):
    '''Like dict.get but looks up keys in multiple dicts.

    :param key: Key to lookup
    :param dicts: Tuple of dicts to use in lookup
    :param default: If passed return this if lookup fails
    '''

    arglen = len(args)
    if arglen not in (2, 3):
        raise ValueError('Must pass at least two args...key and (dicts...)')
    elif arglen == 2:
        key, dicts = args
    else:
        key, dicts, default = args

    for d in dicts:
        value = d.get(key, missing)
        if value is not missing:
            return value

    if 'default' in locals():
        return default

    raise KeyError('{0} does not exist in {1}'.format(key, dicts))


class FieldType(Ordered):
    ''':class:`Form` calls the :meth:`create` to retrieve an appropriate
    control.

    :param nice_name: Nice nice_name of the field (str)
    :param labeled: Field Control has label (bool)
        Overrides the parent Forms labeled attribute for this field only
    :param label_on_top: Label appears on top of the field control (bool)
        Overrides the parent Forms label_on_top attribute for this field only
    :param default: Default value (str)
    '''

    control_cls = None
    control_defaults = None
    field_defaults = {
        'labeled': True,
        'label_on_top': True,
        'default': None,
        'validators': None
    }
    field_keys = ('labeled', 'label_on_top', 'default', 'validators')

    def __init__(self, nice_name, **kwargs):
        super(FieldType, self).__init__()

        self.nice_name = nice_name
        self.control_kwargs = {'name': nice_name}

        # Convert required keyword to required validator
        validators = list(kwargs.pop('validators', []))
        is_required = kwargs.pop('required', False)
        if is_required:
            validators.append(required)
        kwargs['validators'] = validators

        # Set instance attributes from field keys
        for key in self.field_keys:
            set_field_attribute = key in kwargs
            value = kwargs.pop(key, self.field_defaults[key])
            self.control_kwargs[key] = value

            if set_field_attribute:
                setattr(self, key, value)
            else:
                setattr(self, key, None)

        if self.control_defaults:  # If the control has defaults, get em
            for key in self.control_defaults.keys():
                value = kwargs.get(key, self.control_defaults[key])
                self.control_kwargs[key] = value

    def __repr__(self):
        r = '<{}>(nice_name={}, default={})'
        return r.format(self.__class__.__name__, self.nice_name, self.default)

    def create(self):
        return self.control_cls(**self.control_kwargs)


def create_fieldtype(clsname, control_cls, control_defaults=None,
                     field_defaults=None, bases=(FieldType,)):
    '''Convenience function to create a new subclass of :class:`FieldType`.
    *control_defaults* are passed on to *control_cls*. *field_defaults* are
    used as the values of attributes on the returned :class:`FieldType`
    subclass. The keys for both control_defaults and field_defaults are looked
    up in __init__ kwargs param first.

    :param control_cls: PySide widget used to create the control.
    :param control_defaults: Default kwargs to pass to control_cls
    :param field_defaults: Default attr values (labeled, label_on_top, default)

    .. note::

        *control_cls* must  implement the :class:`ControlType` interface
    '''

    bdicts = [b.__dict__ for b in bases]
    orig_defaults = get_key('field_defaults', bdicts)

    if field_defaults:
        defaults = deepcopy(orig_defaults)
        defaults.update(field_defaults)
        field_defaults = defaults
    else:
        field_defaults = orig_defaults

    attrs = {
        'control_cls': control_cls,
        'control_defaults': control_defaults,
        'field_defaults': field_defaults,
    }
    return type(clsname, bases, attrs)


ListField = create_fieldtype(
    'ListField',
    control_cls=controls.ListControl,
    control_defaults={'options': None}
)

BoolField = create_fieldtype(
    'BoolField',
    control_cls=controls.BoolControl,
    field_defaults={'label_on_top': False}
)

StringField = create_fieldtype(
    'StringField',
    control_cls=controls.StringControl,
)

IntField = create_fieldtype(
    'IntField',
    control_cls=controls.IntControl,
    control_defaults={'range': None},
)

FloatField = create_fieldtype(
    'FloatField',
    control_cls=controls.FloatControl,
    control_defaults={'range': None},
)

Int2Field = create_fieldtype(
    'Int2Field',
    control_cls=controls.Int2Control,
    control_defaults={'range1': None, 'range2': None},
)

Float2Field = create_fieldtype(
    'Float2Field',
    control_cls=controls.Float2Control,
    control_defaults={'range1': None, 'range2': None},
)

IntOptionField = create_fieldtype(
    'IntOptionField',
    control_cls=controls.IntOptionControl,
    control_defaults={'options': None},
)

StringOptionField = create_fieldtype(
    'StringOptionField',
    control_cls=controls.StringOptionControl,
    control_defaults={'options': None},
)

ButtonOptionField = create_fieldtype(
    'ButtonOptionField',
    control_cls=controls.ButtonOptionControl,
    control_defaults={'options': None},
)

IntButtonOptionField = create_fieldtype(
    'IntButtonOptionField',
    control_cls=controls.IntButtonOptionControl,
    control_defaults={'options': None},
)

FileField = create_fieldtype(
    'FileField',
    control_cls=controls.FileControl,
    control_defaults={'caption': None, 'filters': None},
)

FolderField = create_fieldtype(
    'FolderField',
    control_cls=controls.FolderControl,
    control_defaults={'caption': None, 'filters': None},
)

SaveFileField = create_fieldtype(
    'SaveFileField',
    control_cls=controls.SaveFileControl,
    control_defaults={'caption': None, 'filters': None},
)

ImageField = create_fieldtype(
    'ImageField',
    control_cls=controls.ImageControl,
)

TextField = create_fieldtype(
    'TextField',
    control_cls=controls.TextControl,
)
InfoField = create_fieldtype(
    'InfoField',
    control_cls=controls.InfoControl,
    control_defaults={'text': None, 'align': 'left'},
    field_defaults={'labeled': False},
)


field_map = {cls.__name__: cls for cls in FieldType.__subclasses__()}
type_map = {
    'image': ImageField,
    'folder': FolderField,
    'text': TextField,
    'file': FileField,
    'str': StringField,
    '(bool,)': ButtonOptionField,
    '(int,)': IntOptionField,
    '(str,)': StringOptionField,
    'int': IntField,
    'float': FloatField,
    '(int, int)': Int2Field,
    '(float, float)': Float2Field,
    'bool': BoolField,
    'list': ListField,
    'info': InfoField,
    str: StringField,
    (bool,): ButtonOptionField,
    (int,): IntOptionField,
    (str,): StringOptionField,
    int: IntField,
    float: FloatField,
    (int, int): Int2Field,
    (float, float): Float2Field,
    bool: BoolField,
    list: ListField,
}



# -*- coding: utf-8 -*-
from __future__ import print_function, absolute_import

import sys

import maya.cmds as cmds
import maya.OpenMayaUI as OpenMayaUI
import maya.OpenMaya as OpenMaya
import maya.utils as utils
from Qt import QtGui, QtCore, QtWidgets

from .renderglobals import RenderGlobals
from .utils import wait, viewport_state, get_maya_window

# Py3 compat
if sys.version_info > (3, 0):
    basestring = str
    long = int


EDITOR_PROPERTIES = [
    'activeComponentsXray',
    'activeOnly',
    'backfaceCulling',
    'bufferMode',
    'bumpResolution',
    'camera',
    'cameras',
    'clipGhosts',
    'colorResolution',
    'controllers',
    'controlVertices',
    'cullingOverride',
    'deformers',
    'depthOfField',
    'dimensions',
    'displayAppearance',
    'displayLights',
    'displayTextures',
    'dynamicConstraints',
    'dynamics',
    'fluids',
    'fogColor',
    'fogDensity',
    'fogEnd',
    'fogging',
    'fogMode',
    'fogSource',
    'fogStart',
    'follicles',
    'gpuCacheDisplayFilter',
    'greasePencils',
    'grid',
    'hairSystems',
    'handles',
    'headsUpDisplay',
    'hulls',
    'ignorePanZoom',
    'ikHandles',
    'imagePlane',
    'interactiveBackFaceCull',
    'interactiveDisableShadows',
    'isFiltered',
    'joints',
    'jointXray',
    'lights',
    'lineWidth',
    'locators',
    'lowQualityLighting',
    'manipulators',
    'maxConstantTransparency',
    'maximumNumHardwareLights',
    'motionTrails',
    'nCloths',
    'nParticles',
    'nRigids',
    'nurbsCurves',
    'nurbsSurfaces',
    'objectFilterShowInHUD',
    'occlusionCulling',
    'particleInstancers',
    'pivots',
    'planes',
    'pluginShapes',
    'polymeshes',
    'rendererName',
    'selectionHiliteDisplay',
    'shadingModel',
    'shadows',
    'smallObjectCulling',
    'smallObjectThreshold',
    'smoothWireframe',
    'sortTransparent',
    'strokes',
    'subdivSurfaces',
    'textureAnisotropic',
    'textureCompression',
    'textureDisplay',
    'textureHilight',
    'textureMaxSize',
    'textures',
    'textureSampling',
    'transparencyAlgorithm',
    'transpInShadows',
    'twoSidedLighting',
    'useBaseRenderer',
    'useDefaultMaterial',
    'useInteractiveMode',
    'useReducedRenderer',
    'viewSelected',
    'wireframeOnShaded',
    'xray',
]

CAMERA_PROPERTIES = [
    'displayFilmGate',
    'displayResolution',
    'displayGateMask',
    'displayFieldChart',
    'displaySafeAction',
    'displaySafeTitle',
    'displayFilmPivot',
    'displayFilmOrigin',
    'overscan',
    'displayGateMaskOpacity',
    'displayGateMaskColor'
]


def deferred_close(view):
    panel = view.panel
    wait(0.1)
    utils.executeDeferred(cmds.deleteUI, panel, panel=True)


def playblast(camera=None, state=None, **kwargs):
    '''Playblast the active viewport.

    Arguments:
        :param camera: Camera to playblast
        :param state: Viewport state

    Playblast Arguments:
        :param width: Resolution width
        :param height: Resolution height
        :param format: Render format like qt or image
        :param compression: Render compression
        :param viewer: Launch the default viewer afterwards
    '''

    playblast_kwargs = {
        'offScreen': False,
        'percent': 100,
        'quality': 100,
        'viewer': False,
        'width': 960,
        'height': 540,
        'framePadding': 4,
        'format': 'qt',
        'compression': 'H.264',
        'forceOverwrite': True,
    }
    playblast_kwargs.update(kwargs)

    active = Viewport.active()
    state = state or active.get_state()

    if camera:
        state['camera'] = camera

    with viewport_state(active, state):
        file = cmds.playblast(**playblast_kwargs)

    return file

class EditorProperty(object):

    def __init__(self, name):
        self.name = name

    def __get__(self, inst, typ=None):
        '''Gets a model editor property.'''

        if not inst:
            return self

        try:
            val = cmds.modelEditor(
                inst.panel,
                **{'query': True, self.name: True}
            )
        except TypeError:
            val = cmds.modelEditor(
                inst.panel,
                **{'query': True, 'qpo': self.name}
            )

        if self.name == 'smallObjectThreshold':
            val = val[0]

        return val

    def __set__(self, inst, value):
        '''Sets a model editor property.'''

        try:
            cmds.modelEditor(
                inst.panel,
                **{'edit': True, self.name: value}
            )
        except TypeError:
            cmds.modelEditor(
                inst.panel,
                **{'edit': True, 'po': [self.name, value]}
            )


class CameraProperty(object):

    def __init__(self, name):
        self.name = name

    def __get__(self, inst, typ=None):
        '''Gets a model panels camera property'''

        if not inst:
            return self

        attr = inst.camera + '.' + self.name
        value = cmds.getAttr(attr)
        if isinstance(value, list):
            if len(value) == 1 and isinstance(value[0], (list, tuple)):
                value = value[0]
        return value

    def __set__(self, inst, value):
        '''Sets a model panels camera property'''

        attr = inst.camera + '.' + self.name

        locked = cmds.getAttr(attr, lock=True)
        if locked:
            return

        has_connections = cmds.listConnections(attr, s=True, d=False)
        if has_connections:
            return

        try:
            if isinstance(value, (int, float)):
                cmds.setAttr(attr, value)
            elif isinstance(value, basestring):
                cmds.setAttr(attr, value, type='string')
            elif isinstance(value, (list, tuple)):
                cmds.setAttr(attr, *value)
            else:
                cmds.setAttr(attr, value)
        except Except as e:
            print('Failed to set state: %s %s' % (attr, value))
            print(e)


class Viewport(object):
    '''A convenient api for manipulating Maya 3D Viewports. While you can
    manually construct a Viewport from an OpenMayaUI.M3dView instance, it is
    much easier to use the convenience methods Viewport.iter,
    Viewport.active and Viewport.get::

        # Get the active view
        v = Viewport.active()
        assert v.focus == True

        # Assuming we have a second modelPanel available
        # Get an inactive view and make it the active view
        v2 = Viewport.get(1)
        v2.focus = True

        assert v.focus == False
        assert v2.focus == True

    Viewport provides standard attribute lookup to all modelEditor properties::

        # Hide nurbsCurves and show polymeshes in the viewport
        v.nurbsCurves = False
        v.polymeshes = True

    :param m3dview: OpenMayaUI.M3dView instance.
    '''

    for p in EDITOR_PROPERTIES:
        locals()[p] = EditorProperty(p)

    for p in CAMERA_PROPERTIES:
        locals()[p] = CameraProperty(p)

    def __init__(self, m3dview):
        self._m3dview = m3dview
        self.highlight = self._highlight
        self.identify = self._highlight

    def __hash__(self):
        return hash(self._m3dview)

    def __eq__(self, other):
        if hasattr(other, '_m3dview'):
            return self._m3dview == other._m3dview
        return self.panel == other

    def copy(self):
        '''Tear off a copy of the viewport.

        :returns: A new torn off copy of Viewport'''

        panel = cmds.modelPanel(tearOffCopy=self.panel)
        view = self.from_panel(panel)
        view.focus = True
        return view

    __copy__ = copy
    __deepcopy__ = copy

    def float(self):
        '''Tear off the panel.'''
        copied_view = self.copy()
        deferred_close(self)
        self._m3dview = copied_view._m3dview

    @classmethod
    def new(cls):
        panel = cmds.modelPanel()
        view = cls.from_panel(panel)
        view.float()
        view.focus = True
        return view

    def close(self):
        '''Close this viewport'''
        deferred_close(self)

    @property
    def properties(self):
        '''A list including all editor property names.'''

        return EDITOR_PROPERTIES + CAMERA_PROPERTIES

    def get_state(self):
        '''Get a state dictionary of all modelEditor properties.'''

        active_state = {}
        active_state['RenderGlobals'] = RenderGlobals.get_state()
        for ep in EDITOR_PROPERTIES + CAMERA_PROPERTIES:
            active_state[ep] = getattr(self, ep)

        return active_state

    def set_state(self, state):
        '''Sets a dictionary of properties all at once.

        :param state: Dictionary including property, value pairs'''

        cstate = state.copy()

        renderglobals_state = cstate.pop('RenderGlobals', None)
        if renderglobals_state:
            RenderGlobals.set_state(renderglobals_state)

        for k, v in cstate.items():
            setattr(self, k, v)

    def playblast(self, camera=None, state=None, **kwargs):
        '''Playblasting with reasonable default arguments. Automatically sets
        this viewport to the active view, ensuring that we playblast the
        correct view.

        :param kwargs: Same kwargs as :func:`maya.cmds.playblast`'''

        if not self.focus:
            self.focus = True

        playblast(camera=None, state=None, **kwargs)

    @property
    def screen_geometry(self):
        qapp = QtWidgets.QApplication.instance()
        desktop = qapp.desktop()
        screen = desktop.screenNumber(self.widget)
        return desktop.screenGeometry(screen)

    def center(self):
        screen_center = self.screen_geometry.center()
        window_center = self.window.rect().center()
        pnt = screen_center - window_center
        self.window.move(pnt)

    @property
    def size(self):
        return self._m3dview.portWidth(), self._m3dview.portHeight()

    @size.setter
    def size(self, wh):
        w1, h1 = self.size
        win_size = self.window.size()
        w2, h2 = win_size.width(), win_size.height()
        w_offset = w2 - w1
        h_offset = h2 - h1
        self.window.resize(wh[0] + w_offset, wh[1] + h_offset)

    @property
    def widget(self):
        '''Returns a QWidget object for the viewport.'''

        try:
            from shiboken import wrapInstance
        except ImportError:  # PySide2 compat
            from shiboken2 import wrapInstance
        w = wrapInstance(long(self._m3dview.widget()), QtWidgets.QWidget)
        return w

    @property
    def window(self):
        '''Returns a QWidget object for the viewports parent window'''

        return self.widget.window()

    @property
    def panel(self):
        '''Returns a panel name for the Viewport.'''
        panel = OpenMayaUI.MQtUtil.fullName(long(self._m3dview.widget()))
        return panel.split('|')[-2]

    @property
    def index(self):
        '''Returns the index of the viewport'''

        i = 0
        for i, view in Viewport.iter():
            if self == view:
                return i
            i += 1
        raise IndexError('Can not find index')

    @property
    def focus(self):
        '''Check if current Viewport is the active Viewport.'''
        return self == self.active()

    @focus.setter
    def focus(self, value):
        '''Set focus to Viewport instance.'''

        if not value:
            try:
                Viewport.get(1).focus = True
            except:
                pass
            return

        cmds.modelEditor(self.panel, edit=True, activeView=True)

    @property
    def camera(self):
        '''Get the short name of the active camera.'''

        camera = OpenMaya.MDagPath()
        self._m3dview.getCamera(camera)
        camera.pop()
        return camera.partialPathName()

    @camera.setter
    def camera(self, camera_path):
        '''Set the active camera for the Viewport.'''

        sel = OpenMaya.MSelectionList()
        sel.add(camera_path)
        camera = OpenMaya.MDagPath()
        sel.getDagPath(0, camera)

        util = OpenMaya.MScriptUtil(0)
        int_ptr = util.asUintPtr()
        camera.numberOfShapesDirectlyBelow(int_ptr)
        num_shapes = util.getUint(int_ptr)
        if num_shapes:
            camera.extendToShape()

        self._m3dview.setCamera(camera)
        self._m3dview.refresh(False, False)

    @property
    def background(self):
        '''Get the background color of the Viewport'''

        color = self._m3dview.backgroundColor()
        return color[0], color[1], color[2]

    @background.setter
    def background(self, values):
        '''Set the background color of the Viewport.

        :param values: RGB value'''

        cmds.displayRGBColor('background', *values)

    def _highlight(self, msec=2000):
        '''Draws an identifier in a Viewport.'''

        highlight = Highlight(self)
        highlight.display(msec)

    @classmethod
    def identify(cls, delay=2000):
        '''Shows identifiers in all Viewports::

            Viewport.identify()

        :param delay: Length of time in ms to leave up identifier
        '''

        cls.highlight()

    @classmethod
    def highlight(cls, msec=2000):
        '''Draws QLabels indexing each Viewport. These indices can be used to
        with :method:`get` to return a corresponding Viewport object.'''

        for viewport in cls.iter():
            if viewport.widget.isVisible():
                viewport.highlight(msec)

    @staticmethod
    def count():
        '''The number of 3D Viewports.'''

        return OpenMayaUI.M3dView.numberOf3dViews()

    @classmethod
    def from_panel(cls, panel):
        m3dview = OpenMayaUI.M3dView()
        OpenMayaUI.M3dView.getM3dViewFromModelPanel(panel, m3dview)
        return cls(m3dview)

    @classmethod
    def get(cls, index):
        '''Get the Viewport at index.'''

        m3dview = OpenMayaUI.M3dView()
        OpenMayaUI.M3dView.get3dView(index, m3dview)
        return cls(m3dview)

    @classmethod
    def active(cls):
        '''Get the active Viewport.'''

        m3dview = OpenMayaUI.M3dView.active3dView()
        return cls(m3dview)

    @classmethod
    def iter(cls):
        '''Yield all Viewport objects.

        usage::

            for view in Viewport.iter():
                print(v.panel)
        '''

        for index in range(cls.count()):
            m3dview = OpenMayaUI.M3dView()
            OpenMayaUI.M3dView.get3dView(index, m3dview)
            yield cls(m3dview)


class Highlight(QtWidgets.QDialog):
    '''Outline a viewport panel and show the panel name.'''

    def __init__(self, view):
        super(Highlight, self).__init__(parent=get_maya_window())
        self.view = view
        self.widget = self.view.widget

        self.setWindowFlags(
            self.windowFlags()
            | QtCore.Qt.FramelessWindowHint
        )
        self.setAttribute(QtCore.Qt.WA_TranslucentBackground)
        self.setAttribute(QtCore.Qt.WA_TransparentForMouseEvents)
        wrect = self.widget.geometry()
        rect = QtCore.QRect(
            self.widget.mapToGlobal(
                wrect.topLeft(),
            ),
            wrect.size(),
        )
        self.setGeometry(
            rect
        )

    def display(self, msec):
        w = QtWidgets.QApplication.instance().activeWindow()
        self.show()
        w.raise_()
        QtCore.QTimer.singleShot(msec, self.accept)

    def paintEvent(self, event):

        painter = QtGui.QPainter(self)

        pen = QtGui.QPen(QtCore.Qt.red)
        pen.setWidth(8)
        font = QtGui.QFont()
        font.setPointSize(48)
        painter.setFont(font)
        painter.setPen(pen)
        painter.setBrush(QtCore.Qt.transparent)

        painter.drawRect(self.rect())
        painter.drawText(self.rect(), QtCore.Qt.AlignCenter, self.view.panel)



# -*- coding: utf-8 -*-

__title__ = 'psforms'
__author__ = 'Dan Bradham'
__email__ = 'danielbradham@gmail.com'
__url__ = 'http://github.com/danbradham/psforms.git'
__version__ = '0.7.0'
__license__ = 'MIT'
__description__ = 'Hassle free PySide forms.'

import os

from . import (controls, exc, fields, resource, widgets)
from .form import Form, FormMetaData
from .validators import *

with open(os.path.join(os.path.dirname(__file__), 'style.css')) as f:
    stylesheet = f.read()



# -*- coding: utf-8 -*-


class Integration(object):

    name = None
    description = None
    icon = None
    banner = None
    requires_confirmation = False
    enabled_by_default = False
    columns = 1

    def __init__(self):
        self.set_enabled(self.enabled_by_default)

    def fields(self):
        '''Return a list of fields.

        Example:
            return [
                {
                    'name': 'StringField',
                    'type': 'str',
                    'default': None,
                    'options': [...],
                    'required': False,
                },
                ...
            ]
        '''
        return NotImplemented

    def on_filename_changed(self, form, value):
        return NotImplemented

    def set_enabled(self, value):
        '''Returns True if the integration was successfully enabled'''

        if value:
            return self._on_enable()
        else:
            return self._on_disable()

    def _on_enable(self):
        self.enabled = self.on_enable()
        return self.enabled

    def on_enable(self):
        '''Return True to enable integration and False to disable'''

        return True

    def _on_disable(self):
        self.enabled = not self.on_disable()
        return self.enabled

    def on_disable(self):
        '''Return True to disable integration and False to enable'''

        return True

    def before_playblast(self, form, data):
        '''Runs before playblasting.'''

        return NotImplemented

    def after_playblast(self, form, data):
        '''Runs after playblasting.'''

        return NotImplemented

    def finalize(self, form, data):
        '''Runs after entire playblast process is finished.

        Unlike after_playblast, this method will only run ONCE after all
        playblasting is finished. So, when playblasting multiple render layers
        you can use this to execute after all of those render layers have
        completed rendering.

        Arguments:
            form: The Form object including render options
            data: List of renders that were output
        '''

        return NotImplemented



# -*- coding: utf-8 -*-

import os
import json
from functools import partial

from maya import cmds, mel

from .forms import PlayblastForm, NewPresetForm, DelPresetForm
from .. import hooks, resources
from ..renderlayers import enabled_render_layers
from ..viewport import playblast, Viewport
from ..utils import get_maya_window
from ..presets import *
from ..vendor.psforms import controls
from ..vendor.psforms.widgets import FormGroup, FormWidget, IconButton
from ..vendor.psforms.form import generate_form
from ..vendor.Qt import QtCore, QtGui, QtWidgets


MISSING = object()
DIALOG_STATE = None
SNAPSHOT_EXT_OPTIONS = {
    'png': '.png',
}
SCENE_STATE_NODE = 'time1'
SCENE_STATE_ATTR = 'mvp_dialog_state'
SCENE_STATE_PATH = SCENE_STATE_NODE + '.' + SCENE_STATE_ATTR
RESOLUTION_OPTIONS = ['Render Settings', '2048x2048', '1920x1080']


def get_sound_track():
    '''Get the current sound track'''

    playback_slider = mel.eval('$tmpVar=$gPlayBackSlider')
    return cmds.timeControl(playback_slider, query=True, sound=True)


def get_fps():
    '''Get the scene fps'''

    # Returns a unit string like game or 5fps
    unit = cmds.currentUnit(query=True, time=True)

    # If unit is not in the following dict - we must have a string like 5fps
    fps = {
        'game': '15fps',
        'film': '24fps',
        'pal': '25fps',
        'ntsc': '30fps',
        'show': '48fps',
        'palf': '50fps',
        'ntscf': '60fps',
    }.get(unit, unit)

    # Convert to a floating point value
    return float(fps.rstrip('fps'))


def get_framerange():
    '''Get the scene framerange'''

    return(
        cmds.playbackOptions(query=True, minTime=True),
        cmds.playbackOptions(query=True, maxTime=True),
    )


class IntegrationUI(object):
    '''Lazily generated UI Group for an Integration class.'''

    def __init__(self, integration, parent=None):
        self.integration = integration
        self.parent = parent
        self._group = None
        self._form = None

    def try_to_toggle(self, value):
        enabled = self.integration.set_enabled(value)

        if enabled and self.group.widget is not self.form:
            self.group.set_widget(self.form)

        self.group.set_enabled(enabled)

    @property
    def group(self):
        if not self._group:
            self._group = FormGroup(
                self.integration.name,
                FormWidget(self.integration.name),  # Stub widget
                parent=self.parent,
            )
            self._group.title.setIcon(QtGui.QIcon(self.integration.icon))
            self._group.toggled.connect(self.try_to_toggle)
            self.try_to_toggle(self.integration.enabled)
        return self._group

    @property
    def form(self):
        if not self._form:
            form = generate_form(
                name=self.integration.name,
                fields=self.integration.fields(),
                title=self.integration.name.title(),
                description=self.integration.description or 'No description',
                icon=self.integration.icon,
                header=False,
                columns=self.integration.columns,
                labels_on_top=False,
            )
            self._form = form.as_widget()

            # Attach controls to integration methods
            for control_name, control in self._form.controls.items():
                if isinstance(control.widget, QtWidgets.QTextEdit):
                    control.widget.setFixedHeight(48)
                method = getattr(
                    self.integration,
                    'on_' + control_name + '_changed',
                    None,
                )

                def slot_method(form, method, control):
                    def slot():
                        method(form, control.get_value())
                    return slot

                if method:
                    control.changed.connect(
                        slot_method(self._form, method, control)
                    )
        return self._form


def get_scene_dialog_state(default=MISSING):
    '''Retrieve dialog state from a maya scene'''

    if cmds.objExists(SCENE_STATE_PATH):
        encoded_state = cmds.getAttr(SCENE_STATE_PATH)
        if encoded_state:
            return json.loads(encoded_state)

    if default is not MISSING:
        return default


def set_scene_dialog_state(state):
    '''Store dialog state in a maya scene'''

    if not cmds.objExists(SCENE_STATE_PATH):
        cmds.addAttr(SCENE_STATE_NODE, ln=SCENE_STATE_ATTR, dataType='string')

    encoded_state = json.dumps(state)
    cmds.setAttr(SCENE_STATE_PATH, encoded_state, type='string')


def new_preset_dialog(parent_dialog):

    dialog = NewPresetForm.as_dialog(parent=parent_dialog)
    dialog.panel.set_options([v.panel for v in Viewport.iter()])

    def on_accept():
        name = dialog.name.get_value()
        panel = dialog.panel.get_value()
        v = Viewport.from_panel(panel)
        state = v.get_state()
        new_preset(name, state)
        update_presets(parent_dialog, name)

    def identify_clicked():
        panel = dialog.panel.get_value()
        v = Viewport.from_panel(panel)
        v.identify()

    identify = QtWidgets.QPushButton('identify')
    identify.clicked.connect(identify_clicked)

    dialog.panel.grid.addWidget(identify, 1, 2)
    dialog.accepted.connect(on_accept)
    dialog.setStyleSheet(resources.get_style())
    dialog.show()


def del_preset_dialog(parent_dialog):

    def on_accept():
        name = parent_dialog.preset.get_value()
        del_preset(name)
        update_presets(parent_dialog)

    dialog = DelPresetForm.as_dialog(parent=parent_dialog)
    dialog.accepted.connect(on_accept)
    dialog.setStyleSheet(resources.get_style())
    dialog.show()


def update_presets(dialog, name=None):

    presets = ['Current Settings'] + sorted([n for n, s in get_presets()])
    dialog.preset.set_options(presets)
    if name:
        dialog.preset.set_value(name)


class PlayblastDialog(object):

    def __init__(self):
        self.form = PlayblastForm.as_dialog(
            frameless=False,
            parent=get_maya_window(),
        )
        self.setup_controls()
        self.setup_connections()
        self.restore_form_state()
        self.apply_stylesheet()

    def show(self):
        self.form.show()

    def store_form_state(self):
        '''Store form state in scene'''
        global DIALOG_STATE
        DIALOG_STATE = self.form.get_value()

        # Update integration ui states
        for name, inst in self.integrations.items():
            if not inst.enabled:
                # Only store integration state for enabled integrations
                DIALOG_STATE.pop(name, None)

        set_scene_dialog_state(DIALOG_STATE)

    def restore_form_state(self):
        '''Restore form state from scene'''
        dialog_state = get_scene_dialog_state(DIALOG_STATE)
        if dialog_state:
            # Update integration ui states
            for name, inst in self.integrations.items():
                state = dialog_state.pop(name, None)
                if state:
                    inst.ui.try_to_toggle(True)
                    inst.ui.group.set_value(**state)
                else:
                    inst.ui.try_to_toggle(False)

            # Update base dialog state
            self.form.set_value(strict=False, **dialog_state)

            if dialog_state['path_option'] != 'Custom':
                self.update_path()

        else:
            # Defaults when no dialog state is present
            self.form.resolution.set_value((
                cmds.getAttr('defaultResolution.width'),
                cmds.getAttr('defaultResolution.height'),
            ))

    def setup_controls(self):
        '''Setup controls and options - Add additional widgets'''

        # Camera options
        cameras = cmds.ls(cameras=True)
        for c in ['frontShape', 'sideShape', 'topShape']:
            if c in cameras:
                cameras.remove(c)
        self.form.camera.set_options(cameras)

        # Viewport Presets
        self.form.preset.grid.setColumnStretch(1, 1)
        presets_menu_button = IconButton(
            icon=resources.get_path('dots-vertical.png'),
            tip='Resolution Presets.',
            name='resolution_menu',
        )
        presets_menu = QtWidgets.QMenu(parent=presets_menu_button)
        presets_menu.addAction('New Preset', partial(new_preset_dialog, self.form))
        presets_menu.addAction('Delete Preset', partial(del_preset_dialog, self.form))
        presets_menu_button.setMenu(presets_menu)
        self.form.preset.grid.addWidget(presets_menu_button, 1, 2)
        update_presets(self.form)

        # Extension options
        ext_option = controls.StringOptionControl(
            'Ext Option',
            labeled=False,
        )
        ext_option.set_options(list(hooks.extension.keys()))
        self.form.controls['ext_option'] = ext_option
        self.form.ext_option = ext_option
        self.form.filename.grid.addWidget(ext_option.widget, 1, 2)

        # Add path option control
        path_options = ['Custom']
        path_options.extend(hooks.pathgen.keys())
        path_option = controls.StringOptionControl(
            'Path Option',
            labeled=False,
        )
        path_option.set_options(path_options)
        self.form.controls['path_option'] = path_option
        self.form.path_option = path_option
        self.form.filename.grid.addWidget(path_option.widget, 1, 3)

        # Identify button
        identify_button = QtWidgets.QPushButton('highlight viewport')
        identify_button.clicked.connect(self.on_identify)
        self.form.button_layout.insertWidget(0, identify_button)

        # Resolution Options
        res_menu_button = IconButton(
            icon=resources.get_path('dots-vertical.png'),
            tip='Resolution Presets.',
            name='resolution_menu',
        )
        res_menu = QtWidgets.QMenu(parent=res_menu_button)
        for option in RESOLUTION_OPTIONS:
            res_menu.addAction(
                option,
                partial(self.apply_resolution_preset, option)
            )
        res_menu_button.setMenu(res_menu)
        self.form.resolution.grid.addWidget(res_menu_button, 1, 2)

        half_res = controls.ToggleControl(
            name='Half Res',
            icon=resources.get_path('half.png'),
            tip='Render half resolution.',
            labeled=False,
        )
        self.form.controls['half_res'] = half_res
        self.form.half_res = half_res
        self.form.resolution.grid.addWidget(half_res.widget, 1, 3)

        # Add postrender hook checkboxes
        self.form.postrender.after_toggled.connect(self.auto_resize)
        self.form.postrender.toggle()
        for postrender in hooks.postrender.values():
            self.form.postrender.add_control(
                postrender.name,
                controls.BoolControl(
                    postrender.name,
                    label_on_top=False,
                    default=postrender.default,
                )
            )

        # Add integration groups
        self.integrations = {}
        for name, integration in hooks.integration.items():
            inst = integration()
            inst.ui = IntegrationUI(inst, self.form)
            inst.ui.group.after_toggled.connect(self.auto_resize)
            inst.form = inst.ui.group
            self.form.add_form(
                name,
                inst.ui.group,
            )
            self.integrations[name] = inst

    def auto_resize(self):
        QtCore.QTimer.singleShot(20, self.form.adjustSize)

    def setup_connections(self):
        '''Connect form controls to callbacks'''

        self.form.capture_mode.changed.connect(self.on_capture_mode_changed)
        self.form.ext_option.changed.connect(self.on_ext_changed)
        self.form.path_option.changed.connect(self.update_path)
        self.form.filename.changed.connect(self.on_filename_changed)
        self.form.accepted.connect(self.on_accept)

    def apply_resolution_preset(self, preset):
        if preset == 'Render Settings':
            self.form.resolution.set_value((
                cmds.getAttr('defaultResolution.width'),
                cmds.getAttr('defaultResolution.height'),
            ))
        else:
            self.form.resolution.set_value([int(num) for num in preset.split('x')])

    def apply_stylesheet(self):
        '''Apply mvp stylesheet'''

        self.form.setStyleSheet(resources.get_style())

    def on_ext_changed(self):
        '''Called when the extension changes.'''
        self.update_path()

    def update_path(self):
        '''Update the file path based on pathgen, ext, and capture mode'''

        path_opt = self.form.path_option.get_value()
        ext_opt = self.form.ext_option.get_value()
        capture_mode = self.form.capture_mode.get_value()

        path_gen = hooks.pathgen.get(path_opt, None)

        if not path_gen:
            return

        # Compute path using PathGenerator hook
        path = path_gen.handler()

        if capture_mode == 'snapshot':

            # Snapshots are just pngs
            ext = SNAPSHOT_EXT_OPTIONS.get(ext_opt, '.png')

        else:

            # Get ext from Extension hook
            extension = hooks.extension.get(ext_opt, None)
            if extension:
                ext = extension.ext
            else:
                ext = '.mov'

            # Put png sequences in a subdirectory
            if ext == '.png':
                scene_path = cmds.file(q=True, shn=True, sn=True)
                scene = os.path.splitext(scene_path)[0]
                path = os.path.join(path, scene).replace('\\', '/')

        # Add extension to path and set filename
        path += ext
        self.form.filename.set_value(path)

        # Notify integrations of filename change
        for integration in self.integrations.values():
            if integration.enabled:
                integration.on_filename_changed(self.form, path)

    def on_capture_mode_changed(self):
        '''Update path when capture_mode changes.'''

        mode = self.form.capture_mode.get_value()
        if mode == 'sequence':
            self.form.ext_option.set_options(list(hooks.extension.keys()))
        else:
            self.form.ext_option.set_options(list(SNAPSHOT_EXT_OPTIONS))

        self.update_path()

    def on_filename_changed(self):
        '''Update contol values when filename changes.'''

        self.form.path_option.set_value('Custom')

        name = self.form.filename.get_value()

        for ext in hooks.extension.values():
            if name.endswith(ext.ext):
                self.form.ext_option.set_value(ext.name)
                return

    def on_identify(self):
        '''Highlight the active viewport.'''

        Viewport.active().identify()

    def on_accept(self):
        '''When form is accepted - parse options and capture'''

        self.store_form_state()
        data = self.form.get_value()

        # Prepare to render
        if data['preset'] == 'Current Settings':
            state = Viewport.active().get_state()
        else:
            state = get_preset(data['preset'])
            state.pop('camera')

        output_dir = os.path.dirname(data['filename'])
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        renders = []
        if data['render_layers'] == 'current':
            renders.append(self._render(state, data))
        else:
            with enabled_render_layers() as layers:
                for layer in layers:
                    # Copy data and add layer name to filename
                    render_data = data.copy()
                    base, ext = render_data['filename'].rsplit('.', 1)
                    layer_name = '{}_{}.{}'.format(base, layer.name(), ext)
                    render_data['filename'] = layer_name

                    layer.switchToLayer()
                    renders.append(self._render(state, render_data))

        # Finalize
        for name, integration in self.integrations.items():
            if integration.enabled:
                integration.finalize(integration.form, renders)

    def _render(self, state, data):

        # Execute integration before_playblast
        for name, integration in self.integrations.items():
            if integration.enabled:
                integration.before_playblast(integration.form, data)

        # Prepare resolution
        if data.pop('half_res', False):
            data['resolution'] = (
                round_to_even(data['resolution'][0] * 0.5),
                round_to_even(data['resolution'][1] * 0.5),
            )

        if data['capture_mode'] == 'snapshot':
            # Render snapshot
            data['start_frame'] = cmds.currentTime(q=True)
            data['end_frame'] = data['start_frame']
            out_file = playblast(
                camera=data['camera'],
                state=state,
                format='image',
                compression='png',
                completeFilename=data['filename'],
                frame=[data['start_frame']],
                width=data['resolution'][0],
                height=data['resolution'][1],
            )
        else:
            # Call extension handler
            extension = hooks.extension.get(data['ext_option'])
            framerange = get_framerange()
            data['start_frame'] = framerange[0]
            data['end_frame'] = framerange[1]
            data['fps'] = get_fps()
            data['sound'] = get_sound_track()
            out_file = extension.handler(
                data=dict(
                    state=state,
                    camera=data['camera'],
                    filename=data['filename'],
                    width=data['resolution'][0],
                    height=data['resolution'][1],
                    sound=data['sound'],
                    fps=data['fps'],
                    start_frame=data['start_frame'],
                    end_frame=data['end_frame'],
                ),
                options=extension.options or {},
            )

        # Execute postrender callbacks
        if 'postrender' in data:
            for name, enabled in data['postrender'].items():
                if enabled:
                    postrender = hooks.postrender.get(name)
                    postrender.handler(data['filename'])

        # Update filename from playblast command
        data['filename'] = out_file

        # Execute integration after_playblast
        for name, integration in self.integrations.items():
            if integration.enabled:
                integration.after_playblast(integration.form, data)

        return data


def round_to_even(value):
    return value if value % 2 == 0 else value + 1


def show():
    '''Main playblast form.'''

    PlayblastDialog._instance = PlayblastDialog()
    PlayblastDialog._instance.show()
    return PlayblastDialog._instance



# -*- coding: utf-8 -*-
'''
Configure mvp presets and hooks
'''

import os
import sys


USER_PRESETS_PATH = os.path.expanduser('~/.mvp')
PRESETS_PATH = [USER_PRESETS_PATH]


def init():
    for path in os.environ.get('MVP_PRESETS', '').split(os.pathsep):
        if path:
            PRESETS_PATH.insert(0, path)

    for path in PRESETS_PATH:

        if not os.path.exists(path):
            os.makedirs(path)

        if path not in sys.path:
            sys.path.insert(1, path)



import os
from . import hooks
from .viewport import playblast


def default_handler(data, options):

    _, ext = os.path.splitext(data['filename'])
    is_qt = ext == '.mov'

    kwargs = dict(
        camera=data['camera'],
        state=data['state'],
        width=data['width'],
        height=data['height'],
    )

    if is_qt:
        kwargs.update(
            format='qt',
            compression='H.264',
            filename=data['filename'],
            sound=data['sound'],
        )
    else:
        # PNG settings
        # No sound
        # .png extension removed
        kwargs.update(
            format='image',
            compression='png',
            filename=data['filename'].rsplit('.', 1)[0],
        )

    return playblast(**kwargs)


hooks.register_extension(
    name='h.264',
    ext='.mov',
    handler=default_handler,
)

hooks.register_extension(
    name='png',
    ext='.png',
    handler=default_handler,
)



'''
General purpose classes and functions.
'''


def itemattrgetter(index, attr):
    '''Returns a function which gets an object in a sequence, then looks up
    an attribute on that object and returns its value.'''

    def get_it(obj):
        return getattr(obj[index], attr)

    return get_it


class Ordered(object):
    '''Maintains the order of creation for all instances/subclasses'''

    _count = 0

    def __init__(self):
        Ordered._count += 1
        self._order = self._count



# -*- coding: utf-8 -*-

# Third party imports
from .. import resources
from ..vendor.psforms import *
from ..vendor.psforms.exc import ValidationError
from ..vendor.psforms.validators import required
from ..vendor.psforms.fields import *


def check_resolution(value):
    if (value[0] % 2) != 0 or (value[1] % 2) != 0:
        raise ValidationError('Resolution must use even numbers.')
    return True


class PostRenderForm(Form):

    meta = FormMetaData(
        title='Post Render',
        description='Post Render Callbacks',
        header=False,
        labels_on_top=False,
        columns=2,
    )


class PlayblastForm(Form):

    meta = FormMetaData(
        header=False,
        icon=resources.get_path('icon.png'),
        title='Review',
        subforms_as_groups=True,
        labels_on_top=False
    )

    filename = SaveFileField(
        'Filename',
        validators=(required,),
    )
    preset = StringOptionField(
        'Viewport Preset',
        options=['Current Settings'],
    )
    capture_mode = StringOptionField(
        'Capture Mode',
        options=['sequence', 'snapshot'],
    )
    render_layers = StringOptionField(
        'Render Layers',
        options=['current', 'all enabled'],
    )
    camera = StringOptionField(
        'Camera',
    )
    resolution = Int2Field(
        'Resolution',
        range1=(0, 8192),
        range2=(0, 8192),
        default=(960, 540),
        validators=(check_resolution,),
    )

    postrender = PostRenderForm()


class NewPresetForm(Form):

    meta = FormMetaData(
        title='New Preset',
        description='Create a new preset from the selected panel.',
        header=False,
    )

    message = InfoField(
        'Message',
        text='Create a new preset from the selected panel.',
        labeled=False,
    )
    panel = StringOptionField('Panel')
    name = StringField('Preset Name', validators=(required,))


class DelPresetForm(Form):

    meta = FormMetaData(
        title='Delete Preset',
        description='Delete the active preset.',
        header=False,
    )

    message = InfoField(
        'Message',
        text='Are you sure you want to delete this preset?',
        labeled=False,
    )



from os.path import abspath, dirname, join
from ..vendor.Qt import QtCore, QtGui


res_package = dirname(__file__)


def get_path(*parts):
    return abspath(join(res_package, *parts)).replace('\\', '/')


def get_qicon(resource, size=None):
    if size:
        pixmap = QtGui.QPixmap(get_path(resource))
        icon = QtGui.QIcon(pixmap.scaled(
            size,
            QtCore.Qt.IgnoreAspectRatio,
            QtCore.Qt.SmoothTransformation,
        ))
    else:
        icon = QtGui.QIcon(get_path(resource))
    return icon


def get_style():
    with open(get_path('style.css'), 'r') as f:
        style = f.read()
    return style



import math

from . import resource
from .exc import *
from .Qt import QtWidgets, QtCore, QtGui


class ControlLayout(QtWidgets.QGridLayout):

    def __init__(self, columns=1, parent=None):
        super(ControlLayout, self).__init__(parent)

        self._columns = columns
        self.setContentsMargins(20, 20, 20, 20)
        self.setHorizontalSpacing(10)
        self.setRowStretch(1000, 1)
        self.widgets = []

    @property
    def columns(self):
        return self._columns

    @columns.setter
    def columns(self, value):
        self._columns = value
        widgets = list(self.widgets)
        for w in widgets:
            self.takeWidget(w)
        for w in widgets:
            self.addWidget(w)

    @property
    def count(self):
        return len(self.widgets)

    def takeWidget(self, widget):
        if widget not in self.widgets:
            return None

        self.widgets.pop(self.widgets.index(widget))
        self.takeAt(self.indexOf(widget))
        return widget

    def addWidget(self, widget):
        count = self.count
        row = math.floor(count / self.columns)
        column = (count % self.columns)
        super(ControlLayout, self).addWidget(widget, row, column)
        self.widgets.append(widget)


class FormWidget(QtWidgets.QWidget):

    def __init__(self, name, columns=1, layout_horizontal=False, parent=None):
        super(FormWidget, self).__init__(parent)

        self.name = name
        self.controls = {}
        self.forms = {}
        self.parent = parent

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.setLayout(self.layout)

        if layout_horizontal:
            self.form_layout = QtWidgets.QBoxLayout(
                QtWidgets.QBoxLayout.LeftToRight
            )
        else:
            self.form_layout = QtWidgets.QBoxLayout(
                QtWidgets.QBoxLayout.TopToBottom
            )
        self.form_layout.setContentsMargins(0, 0, 0, 0)
        self.form_layout.setSpacing(0)

        self.control_layout = ControlLayout(columns=columns)
        self.form_layout.addLayout(self.control_layout)
        self.layout.addLayout(self.form_layout)

        self.setProperty('form', True)
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)

    @property
    def valid(self):
        is_valid = []

        for name, control in self.controls.items():
            control.validate()
            is_valid.append(control.valid)

        for name, form in self.forms.items():
            is_valid.append(form.valid)

        return all(is_valid)

    def get_value(self, flatten=False):
        '''Get the value of this forms fields and subforms fields.

        :param flatten: If set to True, return a flattened dict
        '''

        form_data = {}
        for name, control in self.controls.items():
            form_data[name] = control.get_value()

        for name, form in self.forms.items():
            form_value = form.get_value(flatten=flatten)
            if flatten:
                form_data.update(form_value)
            else:
                form_data[name] = form_value

        return form_data

    def set_value(self, strict=True, **data):
        '''Set the value of all the forms subforms and fields. You can pass
        an additional keyword argument strict to False to ignore mismatched
        names and subforms.

        :param strict: raise exceptions for any invalid names in data
        :param data: Field data used to set the values of the form

        usage::

            myform.set_value(
                strict=True,
                **{
                    'strfield': 'ABCDEFG',
                    'intfield': 1,
                    'subform': {
                        'subform_strfield': 'BCDEFGH',
                        'subform_intfield': 2,}},
            )
        '''
        for name, value in data.items():

            if isinstance(value, dict):
                try:
                    self.forms[name].set_value(**value)
                except KeyError:
                    if strict:
                        raise FormNotFound(name + ' does not exist')
                continue

            try:
                self.controls[name].set_value(value)
            except KeyError:
                if strict:
                    raise FieldNotFound(name + ' does not exist')

    def add_header(self, title, description=None, icon=None):
        '''Add a header'''

        self.header = Header(title, description, icon, self)
        self.layout.insertWidget(0, self.header)

    def insert_form(self, index, name, form):
        '''Insert a subform'''

        self.form_layout.insertWidget(index, form)
        self.forms[name] = form
        setattr(self, name, form)

    def add_form(self, name, form):
        '''Add a subform'''

        self.form_layout.addWidget(form)
        self.forms[name] = form
        setattr(self, name, form)

    def add_control(self, name, control):
        '''Add a control'''

        self.control_layout.addWidget(control.main_widget)
        self.controls[name] = control
        setattr(self, name, control)


class FormDialog(QtWidgets.QDialog):

    def __init__(self, widget, *args, **kwargs):
        super(FormDialog, self).__init__(*args, **kwargs)

        self.widget = widget
        self.cancel_button = QtWidgets.QPushButton('&cancel')
        self.accept_button = QtWidgets.QPushButton('&accept')
        self.cancel_button.clicked.connect(self.reject)
        self.accept_button.clicked.connect(self.on_accept)

        self.layout = QtWidgets.QGridLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setRowStretch(1, 1)
        self.setLayout(self.layout)

        self.button_layout = QtWidgets.QHBoxLayout()
        self.button_layout.setContentsMargins(20, 20, 20, 20)
        self.button_layout.addWidget(self.cancel_button)
        self.button_layout.addWidget(self.accept_button)

        self.layout.addWidget(self.widget, 0, 0)
        self.layout.addLayout(self.button_layout, 2, 0)

    def set_widget(self, widget):
        self.widget = widget
        self.widget.setProperty('groupwidget', True)

        if self.layout.count() == 2:
            self.layout.takeAt(0)

        self.layout.insertWidget(0, self.widget)

    def __getattr__(self, attr):
        try:
            return getattr(self.widget, attr)
        except AttributeError:
            raise AttributeError('FormDialog has no attr: {}'.format(attr))

    def on_accept(self):
        if self.widget.valid:
            self.accept()
        return


class FormGroup(QtWidgets.QWidget):

    toggled = QtCore.Signal(bool)
    after_toggled = QtCore.Signal(bool)

    def __init__(self, name, widget, *args, **kwargs):
        super(FormGroup, self).__init__(*args, **kwargs)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setStretch(1, 0)
        self.layout.setSpacing(0)
        self.setLayout(self.layout)

        self.title = QtWidgets.QPushButton(name)
        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(':/icons/plus'),
            QtGui.QIcon.Normal,
            QtGui.QIcon.Off
        )
        icon.addPixmap(
            QtGui.QPixmap(':/icons/minus'),
            QtGui.QIcon.Normal,
            QtGui.QIcon.On
        )
        self.title.setIcon(icon)
        self.title.setProperty('grouptitle', True)
        self.title.setCheckable(True)
        self.title.setChecked(True)
        self.title.toggled.connect(self.toggle_collapsed)
        self.layout.addWidget(self.title)

        self.widget = widget
        if isinstance(self.widget, QtWidgets.QWidget):
            self.widget.setProperty('groupwidget', True)
            self.layout.addWidget(self.widget)

    def set_widget(self, widget):
        self.widget = widget
        self.widget.setProperty('groupwidget', True)

        if self.layout.count() == 2:
            self.layout.takeAt(1)

        self.layout.addWidget(self.widget)

    def set_enabled(self, value):
        self.title.blockSignals(True)
        self.title.setChecked(value)
        self.widget.setVisible(value)
        self.title.blockSignals(False)

    def toggle_collapsed(self, collapsed):
        self.toggled.emit(collapsed)
        enabled = self.title.isChecked()
        self.widget.setVisible(enabled)
        self.after_toggled.emit(enabled)

    def toggle(self):
        value = not self.title.isChecked()
        self.title.setChecked(value)
        self.toggle_collapsed(value)

    def __getattr__(self, attr):
        try:
            return getattr(self.widget, attr)
        except AttributeError:
            raise AttributeError('FormDialog has no attr: {}'.format(attr))


class Header(QtWidgets.QWidget):

    def __init__(self, title, description=None, icon=None, parent=None):
        super(Header, self).__init__(parent)

        self.grid = QtWidgets.QGridLayout()
        self.grid.setColumnStretch(1, 1)

        self.setLayout(self.grid)

        self.title = QtWidgets.QLabel(title)
        self.title.setProperty('title', True)
        self.descr = QtWidgets.QLabel(description or 'No Description')
        self.descr.setProperty('description', True)
        if not description:
            self.descr.hide()

        if icon:
            self.descr.setAlignment(
                QtCore.Qt.AlignLeft
                | QtCore.Qt.AlignVCenter
            )
            self.title.setAlignment(
                QtCore.Qt.AlignLeft
                | QtCore.Qt.AlignVCenter
            )
            self.icon = QtWidgets.QLabel()
            self.icon.setPixmap(icon)
            if description:
                self.grid.addWidget(self.icon, 0, 0, 2, 1)
                self.grid.addWidget(self.title, 0, 1)
                self.grid.addWidget(self.descr, 1, 1)
            else:
                self.grid.addWidget(self.icon, 0, 0)
                self.grid.addWidget(self.title, 0, 1)
        else:
            self.descr.setAlignment(QtCore.Qt.AlignCenter)
            self.title.setAlignment(QtCore.Qt.AlignCenter)
            self.grid.addWidget(self.title, 0, 0)
            self.grid.addWidget(self.descr, 1, 0)

        self.setProperty('header', True)
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)

        self._mouse_button = None
        self._mouse_last_pos = None

    def mousePressEvent(self, event):
        self._mouse_button = event.button()
        super(Header, self).mousePressEvent(event)
        self._window = self.window()

    def mouseMoveEvent(self, event):
        '''Click + Dragging moves window'''

        if self._mouse_button == QtCore.Qt.LeftButton:
            if self._mouse_last_pos:

                p = self._window.pos()
                v = event.globalPos() - self._mouse_last_pos
                self._window.move(p + v)

            self._mouse_last_pos = event.globalPos()

        super(Header, self).mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        self._mouse_button = None
        self._mouse_last_pos = None
        self._window = None
        super(Header, self).mouseReleaseEvent(event)


class ScalingImage(QtWidgets.QLabel):

    __images = {}

    def __init__(self, image=None, parent=None):
        super(ScalingImage, self).__init__(parent)
        self.images = self.__images
        if not image:
            image = ':/images/noimg'
        self.set_image(image)
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                           QtWidgets.QSizePolicy.Expanding)

    def set_image(self, image):
        if image not in self.images:
            if not isinstance(image, QtGui.QImage):
                if not QtCore.QFile.exists(image):
                    return
                self.img = QtGui.QImage(image)
            self.images[image] = self.img
        else:
            self.img = self.images[image]

        self.setMinimumSize(227, 128)
        self.scale_pixmap()
        self.repaint()

    def scale_pixmap(self):
        scaled_image = self.img.scaled(
            self.width(),
            self.height(),
            QtCore.Qt.KeepAspectRatioByExpanding,
            QtCore.Qt.FastTransformation)
        self.pixmap = QtGui.QPixmap(scaled_image)

    def resizeEvent(self, event):
        self.do_resize = True
        super(ScalingImage, self).resizeEvent(event)

    def paintEvent(self, event):
        if self.do_resize:
            self.scale_pixmap()
            self.do_resize = False

        offsetX = -((self.pixmap.width() - self.width()) * 0.5)
        offsetY = -((self.pixmap.height() - self.height()) * 0.5)
        painter = QtGui.QPainter(self)
        painter.drawPixmap(offsetX, offsetY, self.pixmap)


class IconButton(QtWidgets.QPushButton):
    '''A button with an icon.

    :param icon: path to icon file or resource
    :param tip: tooltip text
    :param name: object name
    :param size: width, height tuple (default: (24, 24))
    '''

    def __init__(self, icon, tip, name, size=(26, 26), *args, **kwargs):
        super(IconButton, self).__init__(*args, **kwargs)

        self.setObjectName(name)
        self.setIcon(QtGui.QIcon(icon))
        self.setIconSize(QtCore.QSize(*size))
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Fixed,
            QtWidgets.QSizePolicy.Fixed)
        self.setFixedHeight(size[0])
        self.setFixedWidth(size[1])
        self.setToolTip(tip)



class FormNotInstantiated(Exception):
    pass

class FieldNotInstantiated(Exception):
    pass

class FieldNotFound(Exception):
    pass

class FormNotFound(Exception):
    pass

class ValidationError(Exception):
    pass



# -*- coding: utf-8 -*-
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from .Qt import QtWidgets, QtCore, QtGui

from .fields import FieldType, type_map, field_map
from .widgets import FormDialog, FormWidget, FormGroup
from .utils import Ordered, itemattrgetter


class FormMetaData(object):

    defaults = dict(
        title='No Title',
        description=None,
        icon=None,
        header=False,
        columns=1,
        labeled=True,
        labels_on_top=True,
        layout_horizontal=False,
        subforms_as_groups=False,
    )

    def __init__(self, **kwargs):
        self.__dict__.update(self.defaults)
        self.__dict__.update(kwargs)


class Form(Ordered):

    meta = FormMetaData()
    _max_width = None

    @classmethod
    def fields(cls):
        '''Returns FieldType objects in sorted order'''

        cls_fields = []
        for name, attr in cls.__dict__.items():
            if issubclass(attr.__class__, FieldType):
                cls_fields.append((name, attr))
        return sorted(cls_fields, key=itemattrgetter(1, '_order'))

    @classmethod
    def forms(cls):
        '''Returns Form objects in sorted order'''

        cls_forms = []
        for name, attr in cls.__dict__.items():
            if issubclass(attr.__class__, Form):
                cls_forms.append((name, attr))
        return sorted(cls_forms, key=itemattrgetter(1, '_order'))

    @classmethod
    def max_width(cls):
        if not cls._max_width:
            # Get the width of the maximum length label
            _max_label = max([y.nice_name for x, y in cls.fields()], key=len)
            _label = QtWidgets.QLabel(_max_label)
            cls._max_width = _label.sizeHint().width() + 20
        return cls._max_width

    @classmethod
    def _create_controls(cls):
        '''Create and return controls from Field objects.'''

        controls = OrderedDict()

        for name, field in cls.fields():
            control = field.create()
            control.setObjectName(name)
            if field.labeled is None:
                control.labeled = cls.meta.labeled
            if field.label_on_top is None:
                control.label_on_top = cls.meta.labels_on_top
            control.label.setFixedWidth(cls.max_width())
            if not control.label_on_top:
                control.errlayout.insertSpacing(0, cls.max_width() + 12)
            controls[name] = control

        return controls

    @classmethod
    def as_widget(cls, parent=None):
        '''Get this form as a widget'''

        form_widget = FormWidget(
            cls.meta.title,
            cls.meta.columns,
            cls.meta.layout_horizontal,
            parent=parent)

        if cls.meta.header:
            form_widget.add_header(
                cls.meta.title,
                cls.meta.description,
                cls.meta.icon
            )

        if cls.fields():
            controls = cls._create_controls()
            for name, control in controls.items():
                form_widget.add_control(name, control)

        for name, form in cls.forms():
            if cls.meta.subforms_as_groups:
                form_widget.add_form(name, form.as_group(form_widget))
            else:
                form_widget.add_form(name, form.as_widget(form_widget))

        return form_widget

    @classmethod
    def as_group(cls, parent=None):

        widget = cls.as_widget()
        group = FormGroup(widget.name, widget, parent=parent)
        return group

    @classmethod
    def as_dialog(cls, frameless=False, dim=False, parent=None):
        '''Get this form as a dialog'''

        dialog = FormDialog(cls.as_widget(), parent=parent)
        dialog.setWindowTitle(cls.meta.title)
        dialog.setWindowIcon(QtGui.QIcon(cls.meta.icon))

        window_flags = dialog.windowFlags()
        if frameless:
            window_flags |= QtCore.Qt.FramelessWindowHint
        dialog.setWindowFlags(window_flags)

        if dim:  # Dim all monitors when showing the dialog
            def _bg_widgets():
                qapp = QtWidgets.QApplication.instance()
                desktop = qapp.desktop()
                screens = desktop.screenCount()
                widgets = []

                for i in range(screens):
                    geo = desktop.screenGeometry(i)
                    w = QtWidgets.QWidget()
                    w.setGeometry(geo)
                    w.setStyleSheet('QWidget {background:black}')
                    w.setWindowOpacity(0.3)
                    widgets.append(w)

                def show():
                    for w in widgets:
                        w.show()

                def hide():
                    for w in widgets:
                        w.hide()

                return show, hide

            old_exec = dialog.exec_
            old_show = dialog.show

            def _exec_(*args, **kwargs):
                bgshow, bghide = _bg_widgets()
                bgshow()
                result = old_exec(*args, **kwargs)
                bghide()
                return result

            def _show(*args, **kwargs):
                bgshow, bghide = _bg_widgets()
                bgshow()
                result = old_show(*args, **kwargs)
                bghide()
                return result

            dialog.exec_ = _exec_
            dialog.show = _show

        return dialog


def generate_form(name, fields, **metadata):
    '''Generate a form from a name and a list of fields.'''

    metadata.setdefault('title', name.title())

    bases = (Form,)
    attrs = {'meta': FormMetaData(**metadata)}
    for field in fields:
        if field['type'] not in type_map:
            raise Exception('Invalid field type %s', field['type'])

        field_type = type_map[field['type']]
        name = field.pop('name')
        label = field.pop('label', name)
        form_field = field_type(label, **field)
        attrs[name] = form_field

    return type(name, bases, attrs)



# -*- coding: utf-8 -*-

import time
from contextlib import contextmanager

from .vendor.Qt import QtWidgets


def get_maya_window(cache=[]):
    '''Get Maya MainWindow as a QWidget.'''

    if cache:
        return cache[0]

    for widget in QtWidgets.QApplication.instance().topLevelWidgets():
        if widget.objectName() == 'MayaWindow':
            cache.append(widget)
            return widget

    raise RuntimeError('Could not locate MayaWindow...')


def wait(delay=1):
    '''Delay python execution for a specified amount of time'''

    app = QtWidgets.QApplication.instance()

    s = time.clock()
    while True:
        if time.clock() - s >= delay:
            return

        app.processEvents()


@contextmanager
def viewport_state(viewport, state):
    '''Sets a viewports state options for the duration of the context.

    Example:

        # Turn off the display of nurbsCurves
        viewport = Viewport.active()
        state = viewport.state()
        state['nurbsCurves'] = False

        with viewport_state(viewport, state):
            # Do something with nurbsCurves off
    '''

    previous_state = viewport.get_state()
    try:
        viewport.set_state(state)
        yield
    finally:
        viewport.set_state(previous_state)



# -*- coding: utf-8 -*-
'''
psforms.controls
================
Wraps standard PySide input widgets providing a unified api for getting and
setting their values. Each control implements :meth:`get_value` and
:meth:`set_value`. A required position argument ``value`` or
``values`` is used to set the default value of the control or in the case
of :class:`ComboBox` and :class:`IntComboBox` a sequence of items to add to
the wrapped QComboBox. In addition each control emits a Signal named `changed`
whenever the value is changed by user interaction.
'''

import os
from copy import deepcopy
from .Qt import QtWidgets, QtCore, QtGui
from . import resource
from .widgets import ScalingImage, IconButton
from .exc import ValidationError


class BaseControl(QtCore.QObject):
    '''Composite Control Object. Used as a base class for all Control Types.
    Subclasses must implement init_widgets, set_value, and get_value methods.
    '''

    changed = QtCore.Signal()
    validate = QtCore.Signal()
    properties = dict(
        valid=True,
    )

    def __init__(self, name, labeled=True, label_on_top=True,
                 default=None, validators=None, *args, **kwargs):
        super(BaseControl, self).__init__(*args, **kwargs)

        self._name = name
        self._labeled = labeled
        self._label_on_top = label_on_top

        self._init_widgets()
        self._init_properties()

        self.validators = validators

        if default:
            self.set_value(default)

    @property
    def name(self):
        '''Property that adjusts the name and label of this control.'''

        return self._name

    @name.setter
    def name(self, value):
        if self.label:
            self.label.setText(self._name)
        self._name = value

    @property
    def labeled(self):
        '''Determines whether this label is visible or hidden.'''

        return self._labeled

    @labeled.setter
    def labeled(self, value):
        self._labeled = value
        if self._labeled:
            self.label.show()
        else:
            self.label.hide()

    @property
    def label_on_top(self):
        '''Determines where the label is drawn, on top or left.'''

        return self._label_on_top

    @label_on_top.setter
    def label_on_top(self, value):
        self._label_on_top = value
        if self._label_on_top:
            self.layout.setDirection(QtWidgets.QBoxLayout.TopToBottom)
        else:
            self.layout.setDirection(QtWidgets.QBoxLayout.LeftToRight)
            self.layout.setSpacing(10)
            if isinstance(self.widget, QtWidgets.QCheckBox):
                self.label.setAlignment(
                    QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter
                )
            else:
                self.label.setAlignment(
                    QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter
                )

    @property
    def valid(self):
        return self.get_property('valid')

    @valid.setter
    def valid(self, value):
        self.set_property('valid', value)

    def validate(self):
        '''Validate this control'''
        if not self.validators:
            return

        value = self.get_value()

        for v in self.validators:
            try:
                v(value)
            except ValidationError as e:
                self.valid = False
                self.errlabel.setText('*' + str(e))
                return

        if not self.valid:
            self.valid = True
            self.errlabel.setText('')

    def emit_changed(self, *args):
        self.changed.emit()
        self.validate()

    def get_property(self, name):
        '''Used to get the value of a property of this control.'''

        return self.properties[name]

    def set_property(self, name, value):
        '''Used to set the value of a property of this control. Consequently
        sets a Qt Property of the same name on all widgets this control
        manages. Allowing the use of these properties in stylesheets.
        '''

        for w in self.widgets:
            w.setProperty(name, value)
        self.properties[name] = value
        self.update_style()

    def update_style(self):
        '''Used to update the style of all widgets this control manages.'''

        for w in self.widgets:
            w.style().unpolish(w)
            w.style().polish(w)

    def init_widgets(self):
        '''Subclasses must implement this method...

        Used to build the widgets for this control.
        Must return a tuple of widgets, where the first element is the main
        widget used to parent this control to a layout. Additionally users
        must attach their widgets to emit the changed signal of this control
        to properly allow for active form validation.
        '''

        raise NotImplementedError()

    def _init_widgets(self):
        '''Binds the widgets returned by init_widgets to self.widget and
        self.widgets.
        '''

        self.widgets = self.init_widgets()
        self.widget = self.widgets[0]

        self.errlabel = QtWidgets.QLabel()
        self.errlabel.setFixedHeight(14)
        self.errlabel.setProperty('err', True)
        self.errlayout = QtWidgets.QHBoxLayout()
        self.errlayout.addWidget(self.errlabel)

        self.label = QtWidgets.QLabel(self.name)
        if isinstance(self.widget, QtWidgets.QCheckBox):
            self.label.setProperty('clickable', True)

            def _mousePressEvent(event):
                self.widget.toggle()
                self.emit_changed()

            self.label.mousePressEvent = _mousePressEvent
            self.errlabel.setAlignment(QtCore.Qt.AlignRight)

        self.widgets = tuple(list(self.widgets) + [self.label])

        if self.label_on_top:
            self.layout = QtWidgets.QBoxLayout(
                QtWidgets.QBoxLayout.TopToBottom
            )
        elif isinstance(self.widget, QtWidgets.QCheckBox):
            self.layout = QtWidgets.QBoxLayout(
                QtWidgets.QBoxLayout.RightToLeft
            )
        else:
            self.label.setAlignment(
                QtCore.Qt.AlignRight |
                QtCore.Qt.AlignVCenter
            )
            self.layout = QtWidgets.QBoxLayout(
                QtWidgets.QBoxLayout.LeftToRight
            )
            self.layout.setSpacing(10)

        self.grid = QtWidgets.QGridLayout()
        self.grid.setContentsMargins(0, 0, 0, 0)
        self.grid.setSpacing(10)
        self.grid.addWidget(self.widget, 1, 1)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self.label)
        self.layout.addLayout(self.grid)
        self.vlayout = QtWidgets.QVBoxLayout()
        self.vlayout.setContentsMargins(0, 0, 0, 0)
        self.vlayout.setSpacing(0)
        self.vlayout.addStretch()
        self.vlayout.addLayout(self.layout)
        self.vlayout.addLayout(self.errlayout)

        if not self.labeled:
            self.label.hide()

        self.main_widget = QtWidgets.QWidget()
        self.main_widget.setLayout(self.vlayout)

    def _init_properties(self):
        '''Initializes the qt properties on all this controls widgets.'''

        self.properties = deepcopy(self.properties)
        for p, v in self.properties.items():
            self.set_property(p, v)

    def get_value(self):
        '''Subclasses must implement this method...

        Must return the value of this control. Use the widgets you create in
        init_widgets, accessible through the widgets attribute after
        initialization.
        '''

        raise NotImplementedError()

    def set_value(self, value):
        '''Subclasses must implement this method...

        Must set the value of this control. Use the widgets you create in
        init_widgets, accessible through the widgets attribute after
        initialization.
        '''

        raise NotImplementedError()


class SpinControl(BaseControl):

    widget_cls = QtWidgets.QSpinBox

    def __init__(self, name, range=None, *args, **kwargs):
        self.range = range
        super(SpinControl, self).__init__(name, *args, **kwargs)

    def init_widgets(self):
        sb = self.widget_cls(parent=self.parent())
        sb.valueChanged.connect(self.emit_changed)
        if self.range:
            sb.setRange(*self.range)
        return (sb,)

    def get_value(self):
        return self.widget.value()

    def set_value(self, value):
        self.widget.setValue(value)


class Spin2Control(BaseControl):

    widget_cls = QtWidgets.QSpinBox

    def __init__(self, name, range1=None, range2=None, *args, **kwargs):
        self.range1 = range1
        self.range2 = range2
        super(Spin2Control, self).__init__(name, *args, **kwargs)

    def init_widgets(self):

        sb1 = self.widget_cls(parent=self.parent())
        sb1.valueChanged.connect(self.emit_changed)
        sb2 = self.widget_cls(parent=self.parent())
        sb2.valueChanged.connect(self.emit_changed)
        if self.range1:
            sb1.setRange(*self.range1)
        if self.range2:
            sb2.setRange(*self.range2)

        w = QtWidgets.QWidget()
        w.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        l = QtWidgets.QHBoxLayout()
        l.setSpacing(10)
        l.setContentsMargins(0, 0, 0, 0)
        l.addWidget(sb1)
        l.addWidget(sb2)
        w.setLayout(l)

        return w, sb1, sb2

    def get_value(self):
        return self.widgets[1].value(), self.widgets[2].value()

    def set_value(self, value):
        self.widgets[1].setValue(value[0])
        self.widgets[2].setValue(value[1])


class IntControl(SpinControl):

    widget_cls = QtWidgets.QSpinBox


class Int2Control(Spin2Control):

    widget_cls = QtWidgets.QSpinBox


class FloatControl(SpinControl):

    widget_cls = QtWidgets.QDoubleSpinBox


class Float2Control(Spin2Control):

    widget_cls = QtWidgets.QDoubleSpinBox


class OptionControl(BaseControl):

    def __init__(self, name, options=None, *args, **kwargs):
        self._init_options = options
        super(OptionControl, self).__init__(name, *args, **kwargs)

    def init_widgets(self):
        c = QtWidgets.QComboBox(parent=self.parent())
        c.activated.connect(self.emit_changed)
        if self._init_options:
            c.addItems(self._init_options)
        return (c,)

    def set_options(self, options):
        self.widget.clear()
        self.widget.addItems(options)

    def get_data(self):
        return self.widget.itemData(
            self.widget.currentIndex(),
            QtCore.Qt.UserRole
        )

    def get_text(self):
        return self.widget.currentText()

    def set_text(self, value):
        self.widget.setCurrentIndex(self.widget.findText(value))

    def get_value(self):
        return self.widget.currentText()

    def set_value(self, value):
        self.widget.setCurrentIndex(self.widget.findText(value))


StringOptionControl = OptionControl


class IntOptionControl(OptionControl):

    def init_widgets(self):

        c = QtWidgets.QComboBox(parent=self.parent())
        c.activated.connect(self.emit_changed)
        return (c,)

    def get_value(self):
        return self.widget.currentIndex()

    def set_value(self, value):
        self.widget.setCurrentIndex(value)


class ButtonOptionControl(BaseControl):

    def __init__(self, name, options, *args, **kwargs):
        self.options = options
        super(ButtonOptionControl, self).__init__(name, *args, **kwargs)

    def init_widgets(self):
        w = QtWidgets.QWidget()
        w.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        l = QtWidgets.QHBoxLayout()
        l.setSpacing(20)
        w.setLayout(l)

        def group_changed(*args):
            self.emit_changed()

        def press_button(index):
            def do_press(*args):
                self.button_group.button(index).setChecked(True)
                self.emit_changed()
            return do_press

        self.button_group = QtWidgets.QButtonGroup()
        self.button_group.buttonClicked.connect(group_changed)

        for i, opt in enumerate(self.options):
            c = QtWidgets.QCheckBox(self.parent())
            if i == 0:
                c.setChecked(True)
            c.setFixedSize(20, 20)

            cl = QtWidgets.QLabel(opt)
            cl.setProperty('clickable', True)
            cl.setAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter)
            cl.mousePressEvent = press_button(i)

            bl = QtWidgets.QHBoxLayout()
            bl.setSpacing(0)
            bl.addWidget(c)
            bl.addWidget(cl)
            self.button_group.addButton(c, i)
            l.addLayout(bl)

        return (w,)

    def get_value(self):
        return self.options[self.button_group.checkedId()]

    def set_value(self, value):

        index = value if isinstance(value, int) else self.options.index(value)
        self.button_group.button(index).setChecked(True)


class IntButtonOptionControl(ButtonOptionControl):

    def get_value(self):
        return self.button_group.checkedId()


class ToggleControl(BaseControl):

    def __init__(self, name, icon, tip=None, *args, **kwargs):
        self.icon = icon
        self.tip = tip
        super(ToggleControl, self).__init__(name, *args, **kwargs)

    def init_widgets(self):
        b = IconButton(self.icon, self.tip, 'checkable')
        b.setCheckable(True)
        b.toggled.connect(self.emit_changed)
        return (b,)

    def get_value(self):
        return self.widget.isChecked()

    def set_value(self, value):
        self.widget.setChecked(value)


class BoolControl(BaseControl):

    def init_widgets(self):
        c = QtWidgets.QCheckBox(parent=self.parent())
        c.setFixedSize(20, 20)
        c.clicked.connect(self.emit_changed)
        return (c, )

    def get_value(self):
        return self.widget.isChecked()

    def set_value(self, value):
        self.widget.setChecked(value)


class StringControl(BaseControl):

    def init_widgets(self):
        le = QtWidgets.QLineEdit(parent=self.parent())
        le.textEdited.connect(self.emit_changed)
        return (le,)

    def get_value(self):
        return self.widget.text()

    def set_value(self, value):
        self.widget.setText(value)


class InfoControl(BaseControl):

    def __init__(self, name, text=None, align='left', *args, **kwargs):
        self.alignment = {
            'left': QtCore.Qt.AlignLeft,
            'center': QtCore.Qt.AlignCenter,
            'left': QtCore.Qt.AlignRight,
        }.get(align.lower(), QtCore.Qt.AlignLeft)

        super(InfoControl, self).__init__(name, *args, **kwargs)

        if text:
            self.set_value(text)


    def init_widgets(self):
        w = QtWidgets.QLabel(alignment=self.alignment, parent=self.parent())
        return (w,)

    def get_value(self):
        return self.widget.text()

    def set_value(self, value):
        self.widget.setText(value)


class TextControl(BaseControl):

    def init_widgets(self):
        te = QtWidgets.QTextEdit(parent=self.parent())
        te.textChanged.connect(self.emit_changed)
        return (te,)

    def get_value(self):
        return self.widget.toPlainText()

    def set_value(self, value):
        self.blockSignals(True)
        self.widget.setText(value)
        self.blockSignals(False)


class BrowseControl(BaseControl):

    browse_method = QtWidgets.QFileDialog.getOpenFileName

    def __init__(self, name, caption=None, filters=None, *args, **kwargs):
        super(BrowseControl, self).__init__(name, *args, **kwargs)
        self.caption = caption or name
        self.filters = filters or ["Any files (*)"]

    def init_widgets(self):

        le = QtWidgets.QLineEdit(parent=self.parent())
        le.setProperty('browse', True)
        le.textEdited.connect(self.emit_changed)
        b = IconButton(
            icon=':/icons/browse_hover',
            tip='Browse',
            name='browse_button',
        )
        b.setProperty('browse', True)
        b.clicked.connect(self.browse)

        w = QtWidgets.QWidget(parent=self.parent())
        l = QtWidgets.QGridLayout()
        l.setContentsMargins(0, 0, 0, 0)
        l.setSpacing(0)
        l.setColumnStretch(0, 1)
        l.addWidget(le, 0, 0)
        l.addWidget(b, 0, 1)
        w.setLayout(l)

        return (w, le, b)

    def get_value(self):
        return self.widgets[1].text()

    def set_value(self, value):
        self.widgets[1].setText(value)

    @property
    def basedir(self):
        line_text = self.get_value()
        if line_text:
            line_dir = os.path.dirname(line_text)
            if os.path.exists(line_dir):
                return line_dir
        return ''

    def browse(self):
        value = self.browse_method(
            self.main_widget,
            caption=self.caption,
            dir=self.basedir)
        if value:
            if self.browse_method is QtWidgets.QFileDialog.getExistingDirectory:
                self.set_value(value)
                return
            self.set_value(value[0])


class FileControl(BrowseControl):

    browse_method = QtWidgets.QFileDialog.getOpenFileName


class FolderControl(BrowseControl):

    browse_method = QtWidgets.QFileDialog.getExistingDirectory


class SaveFileControl(BrowseControl):

    browse_method = QtWidgets.QFileDialog.getSaveFileName


class ImageControl(BaseControl):

    def init_widgets(self):
        w = QtWidgets.QWidget(parent=self.parent())
        i = ScalingImage(parent=w)
        f = FileControl(self.name + '_line', parent=w)
        f.changed.connect(self.emit_changed)
        self.file_control = f

        l = QtWidgets.QVBoxLayout()
        l.setContentsMargins(0, 0, 0, 0)
        l.setSpacing(10)
        l.addWidget(i)
        l.addWidget(f.main_widget)
        w.setLayout(l)

        return [w, i] + list(f.widgets)

    def emit_changed(self, *args):
        self.changed.emit()
        self.widgets[1].set_image(self.get_value())

    def get_value(self):
        return self.file_control.get_value()

    def set_value(self, value):
        if QtCore.QFile.exists(value):
            self.file_control.set_value(value)


class ListControl(BaseControl):

    def __init__(self, name, options=None, *args, **kwargs):
        super(ListControl, self).__init__(name, *args, **kwargs)
        if options:
            self.widget.addItems(options)

    def init_widgets(self):
        l = QtWidgets.QListWidget()
        l.itemSelectionChanged.connect(self.emit_changed)
        return (l,)

    def add_item(self, label, icon=None, data=None):
        item_widget = QtWidgets.QListWidgetItem()
        if icon:
            item_widget.setIcon(QtGui.QIcon(icon))
        if data:
            item_widget.setData(QtCore.Qt.UserRole, data)
        self.widget.addItem(item_widget)

    def get_data(self):
        ''':return: Data for selected items in :class:`QtWidgets.QListWidget`
        :rtype: list'''
        items = self.widget.selectedItems()
        items_data = []
        for item in items:
            items_data.append(item.data(QtCore.Qt.UserRole))
        return items_data

    def get_value(self):
        ''':return: Value of the underlying :class:`QtWidgets.QTreeWidget`
        :rtype: str'''

        items = self.widget.selectedItems()
        item_values = []
        for item in items:
            item_values.append(item.text())
        return item_values

    def set_value(self, value):
        '''Sets the selection of the list to the specified value, label or
        index'''

        if isinstance(value, (str, unicode)):
            items = self.widget.findItems(value)
            if items:
                self.widget.setCurrentItem(items[0])
        elif isinstance(value, int):
            self.widget.setCurrentIndex(int)


control_map = {cls.__name__: cls for cls in BaseControl.__subclasses__()}
